using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using UnityEngine;

/// <summary>
/// This plugin is about body transparency and hiding body parts, while also maintaining functionality of other plugins that manipulate skin material shaders (e.g., tesselation).
/// According to legend, there once was a plugin "Character Transparency" by ZRSX based on Doofenschmalphys edit of Seamonkey's "Hide character" plugin, which itself is a modification of Acidbubbles' "Improved POV".
/// 
/// Version history:
///   0.1: trying to figure out how this works
///   0.2: adding groups
///   0.3: adding selection of parameter to tweak
///   0.4: bugfixes around change of skin
///   0.5: adding special render queues
///   0.6: adding visibility toggles
///   1.0: producing first .var
///   1.1: fixing eye reflection bug
///   1.2: fixing eye reflection render queue order bug, prevent applying Global RQ to special parts, expand help text
///   1.3  (6): fixing UI problem with dropdown list of last item not fitting on screen
///   1.4  (8): adding support for multiple materials of certain body parts, especially male
///   1.5  (9): adding shader replacement for SkinMicroDetail by @northern.shikima
///   1.6 (10): removing the dependency on HuntingSuccubus' skin by having Alpha Mask selection dialogues directly (thanks to incentive and code snippets by AlienAxeMan)
///   1.7 (11): shadows are disabled for transparent objects beyond the shadow cast threshold
/// </summary>

namespace HAL9001
{
	public class GoingGoingGone : MVRScript
	{
		#region PluginInfo    
		
			public string pluginAuthor      = "HAL9001";
			public string pluginName        = "Going Going Gone";
			public string pluginVersion     = "1.7";
			public string pluginDate        = "2024-03-28";
			public string pluginDescription = "Makes body parts transparent or hides them completely";
			
		#endregion

		#region Vars  

			protected Atom                 m_person;
			protected DAZCharacter         m_character;
			protected DAZCharacterSelector m_characterSelector;
			protected DAZSkinV2            m_activeSkin;
		
			protected Dictionary<string, BodyPartControl>  m_controls;
			protected Dictionary<string, BodyGroupControl> m_groupControls;
			
			protected JSONStorableStringChooser m_parameterToEdit;


			protected JSONStorableFloat m_globalTransparency;
			protected JSONStorableFloat m_globalRenderQueue;
			protected JSONStorableBool  m_globalVisibility;
			protected JSONStorableBool  m_globalDebugOutput;
			protected JSONStorableFloat m_shadowCastThreshold;


			protected JSONStorableUrl   m_alphaMaskFace;
			protected JSONStorableUrl   m_alphaMaskTorso;
			protected JSONStorableUrl   m_alphaMaskLimbs;
			protected JSONStorableUrl   m_alphaMaskGens;
			protected Texture[]         m_alphaMaskTextures;

			protected static List<string> sVisibilityChoices = new List<string>  {"On", "Alpha Mask", "Off"};

		#endregion
		
		protected const float DEFAULT_TRANSPARENCY = 0.0f;
		protected const int   DEFAULT_RENDER_QUEUE = 2450;
		
		protected const string PARAMETER_VISIBILITY   = "Visibility";
		protected const string PARAMETER_TRANSPARENCY = "Transparency";
		protected const string PARAMETER_RENDERQUEUE  = "Render Queue";

		protected const int MASK_TEXTURE_IDX_LIMBS = 0;
		protected const int MASK_TEXTURE_IDX_FACE  = 1;
		protected const int MASK_TEXTURE_IDX_TORSO = 2;
		protected const int MASK_TEXTURE_IDX_GENS  = 3;


		protected static readonly string[] BodyMaterialNames = new[]
		{
			"G:Head:Back of the head, face, ears, nostrils & mouth",
				"Head|Back of Head|2",
				"Face||1",
				"Ears||1",
				"Eyelashes||",
				"Irises||",
				"Cornea||",
				"Pupils||",
				"Sclera||",
				"EyeReflection|Eye Reflection|",
				"Lacrimals||",
				"Tear||",
				"Nostrils||1",
				"Lips||1",
				"Gums||",
				"Teeth||",
				"Tongue||",
				"InnerMouth|Inner Mouth|",

			"G:Upper Body:Neck, shoulder, torso & arms",
				"Neck||2",
				"Shoulders||1",
				"Torso||2",
				"Nipples||2",
				"Forearms||0",
				"Hands||0",
				"Fingernails||0",

			"G:Lower Body:Hips, genitals & legs",
				"Hips||2",
				"defaultMat|Genitals Female|3",
				"Genitalia|Genitals Male|3",
				"Genitals||3",
				"Anus|Anus Male|3",
				"Legs||0",
				"Feet||0",
				"Toenails||0",

			// "Hidden",
		};

		protected static readonly Dictionary<string, int> SpecialPartsRenderQueueDefault = new Dictionary<string, int>
		{
			{ "Tear",           3000 },
			{ "Eye Reflection", DEFAULT_RENDER_QUEUE + 10 }
		};

		protected static readonly Dictionary<string, Shader> TransparencyReplacementShaders = new Dictionary<string, Shader>
		{
			// Opaque materials
			{ "Custom/Subsurface/GlossCullComputeBuff",               Shader.Find("Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/GlossNMCullComputeBuff",             Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/GlossNMDetailCullComputeBuff",       Shader.Find("Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/CullComputeBuff",                    Shader.Find("Custom/Subsurface/TransparentSeparateAlphaComputeBuff") },
			// Hunting-Succubus' tesselation material
            { "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff",  Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
            // NorthernShikima.SkinMicroDetail tesselation material
            { "Custom/Subsurface/GlossNMDetailTessMappedComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
			// Transparent materials (don't need a replacement, but can be "driven" with "AlphaAdjust")
			{ "Custom/Subsurface/TransparentComputeBuff",                                 null },
			{ "Custom/Subsurface/TransparentSeparateAlphaComputeBuff",                    null },
			{ "Custom/Subsurface/TransparentGlossComputeBuff",                            null },
			{ "Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff",               null },
			{ "Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff",             null },
			{ "Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff", null },
			{ "Custom/Subsurface/TransparentGlossNoCullSeparateAlphaComputeBuff",         null },
			{ "Custom/Subsurface/AlphaMaskComputeBuff",                                   null },
			{ "Marmoset/Transparent/Simple Glass/Specular IBLComputeBuff",                null },
		};


		protected class MaterialSlot
		{
			protected GoingGoingGone parentScript;
			protected string         bodyPart;
			protected int            materialSlot;
			protected int            skinArrayIndex;
			protected Material       material;
			protected Shader         originalShader, replacementShader;
			protected float          originalAlphaAdjust, originalAlphaCutoff;
			protected Color          originalColor;
			protected Color          originalSpecular;
			protected int            originalQueue;
			protected Texture        originalAlphaTexture;
			
			public MaterialSlot(GoingGoingGone _parent, Material _material, string _bodyPart, int _slotIdx, int _skinArrayIdx)
			{
				parentScript      = _parent;
				bodyPart          = _bodyPart;
				materialSlot      = _slotIdx;		
				skinArrayIndex    = _skinArrayIdx;		
				material          = _material;
				originalShader    = null;
				replacementShader = null;

				if (parentScript.m_globalDebugOutput.val)
				{
					StringBuilder sb = new StringBuilder();
					for ( int i=0; i<material.passCount; i++)
					{
						if (i>0) sb.Append(", ");
						sb.Append($"{i}:'{material.GetPassName(i)}'");
					}
					parentScript.logDebug($"{bodyPart}#{materialSlot} '{material.name}': Shader={_material.shader.name}, RQ={material.renderQueue}, Passes=[{sb.ToString()}]");
				}
			}
			
			public bool HasShaderChanged()
			{
				// something changed the shader from outside
				return (replacementShader != null) && (material.shader != replacementShader);
			}
			
			public bool ReplaceShader()
			{
				bool shaderReplaced = false;
				if ((replacementShader == null) || HasShaderChanged())
				{
					// only replace when not done yet or something happened to the shader from outside
					if (TransparencyReplacementShaders.TryGetValue(material.shader.name, out replacementShader))
					{
						originalShader       = material.shader;
						originalQueue        = material.renderQueue;
						originalAlphaAdjust  = material.GetFloat(  "_AlphaAdjust");
						originalAlphaTexture = material.GetTexture("_AlphaTex"   );
						originalAlphaCutoff  = material.GetFloat(  "_Cutoff"     ); 
						originalColor        = material.GetColor(  "_Color"      );
						originalSpecular     = material.GetColor(  "_SpecColor"  );

						if (replacementShader != null)
						{
							material.shader = replacementShader;
							shaderReplaced  = true;
												
							parentScript.logDebug($"{bodyPart}#{materialSlot} '{material.name}' replacing shader with {replacementShader.name}");
						}
						else
						{
							// "null" as replacement shader means that the material already has a transparent capable shader
							// "keep" original shader
							replacementShader = material.shader;
							
							parentScript.logDebug($"{bodyPart}#{materialSlot} '{material.name}' keeping shader {replacementShader.name}");
						}
					}
				}
				return shaderReplaced;
			}
			
			public bool RestoreShader()
			{
				bool shaderRestored = false;
				if (replacementShader != null)
				{				
					material.renderQueue = originalQueue;
					material.shader      = originalShader;
					material.SetFloat(  "_AlphaAdjust", originalAlphaAdjust);
					material.SetTexture("_AlphaTex",    originalAlphaTexture);
					material.SetFloat(  "_Cutoff",      originalAlphaCutoff);
					material.SetColor(  "_Color",       originalColor);
					material.SetColor(  "_SpecColor",   originalSpecular);
					
					if (replacementShader != originalShader)
					{
						parentScript.logDebug($"{bodyPart}#{materialSlot} '{material.name}' restored shader {originalShader.name}");
					}

					originalShader    = null; // restore has happened
					replacementShader = null;
					shaderRestored    = true;
				}
				return shaderRestored;
			}
			
			public void SetMaterialValues(int _renderQueue, float _transparency, int _alphaTextureIdx)
			{
				if (replacementShader == null) return; // don't apply when not replaced
				
				material.renderQueue = _renderQueue;
				
				material.SetFloat("_AlphaAdjust", Mathf.Min(originalAlphaAdjust, -_transparency));
				
				Color col = originalColor;
				col.a  = Mathf.Min(col.a, 1.0f - _transparency);
				material.SetColor("_Color", col);
				
				col    = originalSpecular;
				col.a  = Mathf.Min(col.a, 1.0f - _transparency);
				if (bodyPart == "Eye Reflection")
				{
					// treat differently to avoid reflection showing up for too long
					col.r *= col.a;
					col.g *= col.a;
					col.b *= col.a;
				}
				material.SetColor("_SpecColor", col);

				if (_alphaTextureIdx >= 0)
				{
					material.SetTexture("_AlphaTex", parentScript.m_alphaMaskTextures[_alphaTextureIdx]);					
					material.SetFloat(  "_Cutoff",   0.5f - (_transparency * 0.5f));
				}
				else
				{
					material.SetTexture("_AlphaTex", originalAlphaTexture);					
					material.SetFloat(  "_Cutoff",   originalAlphaCutoff);
				}

				// getting rid of shadows when totatally transparent
				parentScript.m_activeSkin.materialsShadowCastEnabled[skinArrayIndex] = _transparency < parentScript.m_shadowCastThreshold.val;
			}
		}


		protected class BodyPartControl
		{
			protected GoingGoingGone            parentScript;
			protected string                    storableName;
			protected int                       alphaTextureIdx;
			
			protected JSONStorableStringChooser currVisibility;
			protected JSONStorableFloat         currTransparency;
			protected JSONStorableFloat         currRenderQueue;
			
			protected bool                      oldSetAlphaMask;
			protected float                     oldTransparency;
			protected int                       oldRenderQueue;

			protected List<MaterialSlot>        materials;
			protected bool                      replacementShadersActive;

			
			public BodyPartControl(GoingGoingGone _parent, string _storableName, int _alphaTextureIdx)
			{
				parentScript    = _parent;
				storableName    = _storableName;
				alphaTextureIdx = _alphaTextureIdx;

				currVisibility = new JSONStorableStringChooser(storableName + " Visibility", sVisibilityChoices, sVisibilityChoices[0], storableName + " Visibility");
				parentScript.RegisterStringChooser(currVisibility);

				currTransparency = new JSONStorableFloat(storableName + " Transparency", DEFAULT_TRANSPARENCY, 0f, 1f, true);
				parentScript.RegisterFloat(currTransparency);
				
				int defaultRQ;
				if (!SpecialPartsRenderQueueDefault.TryGetValue(storableName, out defaultRQ)) { defaultRQ = DEFAULT_RENDER_QUEUE; }
				currRenderQueue = new JSONStorableFloat(storableName + " Render Queue", defaultRQ, 0f, 5000f, true);
				parentScript.RegisterFloat(currRenderQueue);
				
				oldTransparency = -1.0f; // force update ASAP
				oldSetAlphaMask = false;
				oldRenderQueue  = -1;
				
				materials                = new List<MaterialSlot>();
				replacementShadersActive = false;
			}

			public void CreateUI()
			{
				switch (parentScript.m_parameterToEdit.val)
				{
					case PARAMETER_VISIBILITY :
					{
						var selVS = parentScript.CreatePopup(currVisibility, false);
						selVS.label = storableName;
						break;
					}

					case PARAMETER_TRANSPARENCY : 
					{
						var sldT = parentScript.CreateSlider(currTransparency, false);
						sldT.label              = storableName;
						sldT.rangeAdjustEnabled = false;
						break;
					}
					
					case PARAMETER_RENDERQUEUE :
					{
						var sldRQ = parentScript.CreateSlider(currRenderQueue, false);
						sldRQ.valueFormat = "F0";
						break;
					}
				}
			}

			public void DestroyUI()
			{
				parentScript.RemovePopup( currVisibility);
				parentScript.RemoveSlider(currTransparency);
				parentScript.RemoveSlider(currRenderQueue);
			}

			public void Reset()
			{
				RestoreShaders();
				materials.Clear();
				ForceUpdate();
			}

			public void AddMaterial(Material _material, int skinArrayIndex)
			{
				if (_material != null)
				{
					MaterialSlot ms = new MaterialSlot(parentScript, _material, storableName, materials.Count+1, skinArrayIndex);
					materials.Add(ms);
					ForceUpdate();
				}					
				
			}

			public bool ApplyTransparency()
			{
				if (materials.Count == 0) return false;

				bool  shadersChanged           = false;
				bool  shadersChangedExternally = materials.Any(m => m.HasShaderChanged()); // something changed the replacement shader externally
				float transparency             = currTransparency.val;
				int   renderQueue              = (int) currRenderQueue.val;

				bool setAlphaMask = false;
				if (currVisibility.val == sVisibilityChoices[1]) 
				{ 
					 // AlphaMasked > enforce shader swap
					transparency = Mathf.Max(0.001f, transparency);
					setAlphaMask = true;
				}
				else if (currVisibility.val == sVisibilityChoices[2]) 
				{ 
					// Off
					transparency = 1.0f;
				} 

				if (shadersChangedExternally || (transparency != oldTransparency) || (renderQueue != oldRenderQueue) || (setAlphaMask != oldSetAlphaMask))
				{
					if (transparency <= 0.0f)
					{
						// no transparency > restore original shaders
						if (replacementShadersActive)
						{
							shadersChanged |= RestoreShaders();
							replacementShadersActive = false;
						}
					}
					else
					{
						if (!replacementShadersActive || shadersChangedExternally)
						{
							// transparency needed (or shader changed) > shader replacement necessary
							bool replaced = ReplaceShaders();
							shadersChanged |= replaced;
							replacementShadersActive = true;
						}
					}

					if (replacementShadersActive)
					{
						// apply alpha texture (if exists)
						foreach (MaterialSlot ms in materials)
						{
							ms.SetMaterialValues(renderQueue, transparency, setAlphaMask ? alphaTextureIdx : -1);
						}
					}

					oldTransparency = transparency;
					oldRenderQueue  = renderQueue;
					oldSetAlphaMask = setAlphaMask;
				}
				return shadersChanged;
			}

			public bool ReplaceShaders()
			{
				bool shadersReplaced = false;
				foreach (MaterialSlot ms in materials) 
				{
					shadersReplaced |= ms.ReplaceShader();
				}
				return shadersReplaced;
			}

			public bool RestoreShaders()
			{
				bool shadersRestored = false;
				if (replacementShadersActive)
				{
					foreach (MaterialSlot ms in materials) 
					{
						shadersRestored |= ms.RestoreShader();
					}
					replacementShadersActive = false;
				}
				return shadersRestored;					
			}

			public void SetVisibility(string _choice)
			{
				currVisibility.val = _choice;
			}

			public void SetTransparency(float _transparency)
			{
				currTransparency.val = _transparency;
			}

			public void SetRenderQueue(int _renderQueue)
			{
				currRenderQueue.val = _renderQueue;
			}

			public void ForceUpdate()
			{
				oldTransparency = -1;
			}

			public bool UsesSpecificRenderQueue()
			{
				return SpecialPartsRenderQueueDefault.ContainsKey(storableName);
			}
		}


		protected class BodyGroupControl
		{
			protected GoingGoingGone            parentScript;
			protected string                    groupName;

			protected JSONStorableString        groupHeader;
			protected JSONStorableFloat         groupTransparency;
			protected JSONStorableStringChooser groupVisibility;
			protected JSONStorableBool          groupVisible;
			protected JSONStorableBool          groupExpanded;
			protected UIDynamic			        bottomSpacer;

			protected List<BodyPartControl> controls;

			public BodyGroupControl(GoingGoingGone _parent, string _groupName, string _description)
			{
				parentScript = _parent;
				groupName    = _groupName;

				groupHeader = new JSONStorableString(groupName, "<color=#000><size=32><b>" + groupName + "</b></size></color>\n" + _description);

				groupExpanded = new JSONStorableBool("Expand Group", false, (bool _val) => parentScript.RefreshUI());

				groupTransparency = new JSONStorableFloat(_groupName + " Group Transparency", 0f, (_val) => ApplyGroupTransparency(_val), 0f, 1f, true);
				parentScript.RegisterFloat(groupTransparency);
				groupTransparency.isStorable   = false;
				groupTransparency.isRestorable = false;

				groupVisibility = new JSONStorableStringChooser(_groupName + " Group Visibility", sVisibilityChoices, sVisibilityChoices[0], _groupName + " Group Visibility", (_val) => ApplyGroupVisibility(_val));
				parentScript.RegisterStringChooser(groupVisibility);
				groupVisibility.isStorable   = false;
				groupVisibility.isRestorable = false;
				
				groupVisible = new JSONStorableBool(_groupName + " Group Visible", true, (_val) => ApplyGroupVisible(_val));
				parentScript.RegisterBool(groupVisible);
				groupVisible.isStorable   = false;
				groupVisible.isRestorable = false;

				controls     = new List<BodyPartControl>();
				bottomSpacer = null;
			}

			public void AddControl(BodyPartControl _control)
			{
				controls.Add(_control);
			}

			public void CreateUI()
			{
				var txtGroupHeader = parentScript.CreateTextField(groupHeader, false);
				txtGroupHeader.height = 50.0f;

				var tglExpand = parentScript.CreateToggle(groupExpanded, false);

				var sldT   = parentScript.CreateSlider(groupTransparency, false);
				sldT.label = "Group Transparency";

				var togVS   = parentScript.CreateToggle(groupVisible, false);
				togVS.label = "Group Visible";
						
				if (groupExpanded.val == true)
				{
					foreach (BodyPartControl control in controls) 
					{
						control.CreateUI();
					}					
				}
				
				// sadly necessary UI hack to give the last popup at the bottom of the screen space to expand
				if ((groupName == "Lower Body") && (parentScript.m_parameterToEdit.val == PARAMETER_VISIBILITY))
				{
					bottomSpacer = parentScript.CreateSpacer(false);
					bottomSpacer.height = 150f;
				}				
			}

			public void DestroyUI()
			{
				foreach (BodyPartControl control in controls) 
				{
					control.DestroyUI();
				}
				parentScript.RemoveSlider(groupTransparency);
				parentScript.RemoveToggle(groupVisible);
				parentScript.RemoveToggle(groupExpanded);
				parentScript.RemoveTextField(groupHeader);
				if (bottomSpacer != null)
				{
					parentScript.RemoveSpacer(bottomSpacer);
					bottomSpacer = null;
				}					
			}

			public void SetTransparency(float _value)
			{
				groupTransparency.val = _value;
			}

			public void SetVisibility(bool _value)
			{
				groupVisible.val = _value;
			}

			protected void ApplyGroupTransparency(float _value)
			{
				foreach (var control in controls)
				{
					control.SetTransparency(_value);
				}
			}

			public void ApplyGroupVisibility(string _value)
			{
				foreach (var control in controls)
				{
					control.SetVisibility(_value);
				}
			}

			public void ApplyGroupVisible(bool _value)
			{
				ApplyGroupVisibility(_value ? sVisibilityChoices[0] : sVisibilityChoices[2]);
			}
		}

		
		public override void Init()
		{
			try
			{
				if (containingAtom?.type != "Person")
				{
					logError($"This plugin only works on Person atoms");
					DestroyImmediate(this);
					return;
				}

				MacGruber.Utils.OnInitUI(CreateUIElement);

				m_person            = containingAtom;
				m_characterSelector = m_person.GetComponentInChildren<DAZCharacterSelector>();
				m_character         = null;
				m_activeSkin        = null;

				m_alphaMaskTextures = new Texture[4];
				for (int i=0; i<m_alphaMaskTextures.Length; i++)
				{
					m_alphaMaskTextures[i] = Texture2D.whiteTexture; // null is not an option
				}

				// Help text
				MacGruber.Utils.SetupInfoText(this,
					"<color=#000><size=32><b>Going Going Gone</b></size></color>\n\n" +
					"<color=#333>" +
					"This plugin changes the visibility and transparency of body parts.\n\n" +
					"Moving the sliders will change the transparency of the related body part.\n\n" +
					"'Parameter to Edit' determines which parameter (Visibility, Transparency, Render Queue) is shown in the UI\n\n" +
					"The controls are split into three groups (Head, Upper Body, Lower Body). Check the 'Expand Group' toggles to see all controllers of that group.\n\n" +
					"For quickly toggling the visibility of all body parts, use the 'Global Visibility' toggle. Be aware that it will change all visibility values.\n\n" +
					"For quickly applying the same transparency to all body parts, use the 'Global Transparency' slider. Be aware that it will change all sliders.\n\n" +
					"For quickly applying the same render queue to all body parts, use the 'Global Render Queue' slider. Be aware that it will change all sliders (except for 'Eye Reflection' and 'Tears')." +
					"</color>", 
					250, true);

				m_parameterToEdit = new JSONStorableStringChooser("Parameter to Edit", 
					new List<string>{ PARAMETER_VISIBILITY, PARAMETER_TRANSPARENCY, PARAMETER_RENDERQUEUE },
					PARAMETER_VISIBILITY, "Parameter to Edit",
					(string _val) => RefreshUI()
				);
				m_parameterToEdit.isStorable = false;
				CreatePopup(m_parameterToEdit, true);

				m_globalVisibility = new JSONStorableBool("Global Visibility", true, (_val) => ApplyVisibilityToAll(_val));
				RegisterBool(m_globalVisibility);
				m_globalVisibility.isStorable   = false;
				m_globalVisibility.isRestorable = false;
				CreateToggle(m_globalVisibility, true);

				m_globalTransparency = new JSONStorableFloat("Global Transparency", DEFAULT_TRANSPARENCY, (_val) => ApplyTransparencyToAll(_val), 0f, 1f, true);
				RegisterFloat(m_globalTransparency);
				m_globalTransparency.isStorable   = false;
				m_globalTransparency.isRestorable = false;
				var sliderT = CreateSlider(m_globalTransparency, true);

				m_globalRenderQueue = new JSONStorableFloat("Global Render Queue", DEFAULT_RENDER_QUEUE, (_val) => ApplyRenderQueueToAll(_val), 0, 5000, true);
				RegisterFloat(m_globalRenderQueue);
				m_globalRenderQueue.isStorable   = false;
				m_globalRenderQueue.isRestorable = false;
				var sliderRQ = CreateSlider(m_globalRenderQueue, true);
				sliderRQ.valueFormat = "F0";
				
				m_shadowCastThreshold = new JSONStorableFloat("Shadow Cast Threshold", 0.9f, 0f, 1f);
				CreateSlider(m_shadowCastThreshold, true);
				RegisterFloat(m_shadowCastThreshold);

				m_globalDebugOutput = new JSONStorableBool("Debug Output", false);
				RegisterBool(m_globalDebugOutput);
				var tglDebug = CreateToggle(m_globalDebugOutput, true);

                MacGruber.Utils.SetupSpacer(this, 20, true);
				MacGruber.Utils.SetupInfoText(this, 
                    "<color=#000><size=32><b>Alpha Textures</b></size></color>",
					50, true);
                MacGruber.TextureSettings settings = new MacGruber.TextureSettings();
				settings.createAlphaFromGrayscale = true;
                m_alphaMaskFace  = MacGruber.Utils.SetupTexture2DChooser(this, "Face Alpha Mask",  "None", true, settings, (_tex) => OnSetAlphaMaskTexture(_tex, MASK_TEXTURE_IDX_FACE));
                m_alphaMaskTorso = MacGruber.Utils.SetupTexture2DChooser(this, "Torso Alpha Mask", "None", true, settings, (_tex) => OnSetAlphaMaskTexture(_tex, MASK_TEXTURE_IDX_TORSO));
                m_alphaMaskLimbs = MacGruber.Utils.SetupTexture2DChooser(this, "Limbs Alpha Mask", "None", true, settings, (_tex) => OnSetAlphaMaskTexture(_tex, MASK_TEXTURE_IDX_LIMBS));
                m_alphaMaskGens  = MacGruber.Utils.SetupTexture2DChooser(this, "Gens Alpha Mask",  "None", true, settings, (_tex) => OnSetAlphaMaskTexture(_tex, MASK_TEXTURE_IDX_GENS));

				m_controls      = new Dictionary<string, BodyPartControl>();
				m_groupControls = new Dictionary<string, BodyGroupControl>();
				BodyGroupControl activeGroup = null;
				foreach (string bodyMaterialName in BodyMaterialNames)
				{
					if (bodyMaterialName.StartsWith("G:"))
					{
						// control group
						string[] parts       = bodyMaterialName.Split(':');
						string   title       = parts[1];
						string   description = parts[2];
						
						activeGroup = new BodyGroupControl(this, title, description);
						m_groupControls[title] = activeGroup;
					}
					else
					{
						// "normal" part
						string[] parts        = bodyMaterialName.Split('|');
						string   materialName = parts[0];
						string   storableName = parts[1].Length == 0 ? materialName : parts[1];
						int      alphaTexIdx  = parts[2].Length == 0 ? -1 : Int32.Parse(parts[2]);

						var control = new BodyPartControl(this, storableName, alphaTexIdx);
						m_controls[materialName] = control;
						activeGroup.AddControl(control);
					}					
				}

				RefreshUI();
			}
			catch (Exception _cause)
			{
				logError("Initialisation error " + _cause);
				DestroyImmediate(this);
			}
		}

		
        public void OnSetAlphaMaskTexture(Texture2D tex, int textureID)
        {
			// need to clean up any previous texture
			if (m_alphaMaskTextures[textureID] != Texture2D.whiteTexture)
			{
				Destroy(m_alphaMaskTextures[textureID]);
				m_alphaMaskTextures[textureID] = Texture2D.whiteTexture;
			}

            if (tex != null) 
			{
				m_alphaMaskTextures[textureID] = tex;
				logDebug($"OnSetAlphaMaskTexture({textureID}): {tex.name} [{tex.format}] ({tex.width}, {tex.height})");
			}
			else
			{
				logDebug($"OnSetAlphaMaskTexture({textureID}): None");
			}

            ForceUpdate();
        }


		public void RefreshUI()
		{
			// brute force approach > delete all controls on the left
			foreach (var group in m_groupControls)
			{
				group.Value.DestroyUI();
			}
			// then recreate them 
			foreach (var group in m_groupControls)
			{
				group.Value.CreateUI();
			}
		}

		protected void ApplyVisibilityToAll(bool _value)
		{
			foreach (var group in m_groupControls)
			{
				group.Value.SetVisibility(_value);
			}
		}

		protected void ApplyTransparencyToAll(float _value)
		{
			foreach (var group in m_groupControls)
			{
				group.Value.SetTransparency(_value);
			}
		}

		protected void ApplyRenderQueueToAll(float _value)
		{
			foreach (var control in m_controls)
			{
				if (!control.Value.UsesSpecificRenderQueue())
				{
					control.Value.SetRenderQueue((int) _value);
				}
			}
		}

		public void OnEnable()
		{
			ForceUpdate();
		}
		
		public void OnDisable()
		{
			RestoreShaders();
		}
		
		public void OnDestroy()
		{
			OnDisable();
			// destroy loaded textures
			for (int i=0; i < m_alphaMaskTextures.Length; i++)
			{
				if (m_alphaMaskTextures[i] != Texture2D.whiteTexture)
				{
					Destroy(m_alphaMaskTextures[i]);
					m_alphaMaskTextures[i] = Texture2D.whiteTexture;
				}
			}
			MacGruber.Utils.OnDestroyUI();
		}

		public void Update()
		{
			if (m_characterSelector.selectedCharacter != m_character)
			{
				RestoreShaders();
				m_activeSkin = null;
				m_character  = m_characterSelector.selectedCharacter;
			}

			if (m_character != null)
			{
				bool shaderChanged = false;
				if (m_activeSkin != m_character.skin)
				{
					logDebug($"Skin change detected ({m_character.skin.sceneGeometryId})");
					
					// skin changed > reset controls
					foreach (var control in m_controls) 
					{
						control.Value.Reset();
					}
					
					m_activeSkin = m_character.skin;
					if (!m_activeSkin.GPUuseSimpleMaterial)
					{
						for (int idx = 0; idx < m_activeSkin.GPUmaterials.Length ; idx++ )
						{
							Material material = m_activeSkin.GPUmaterials[idx];
							string partName = material.name.Split('-')[0]; // remove "-1/2/3..." postfix used for multiple characters
							BodyPartControl control;
							if (m_controls.TryGetValue(partName, out control))
							{
								control.AddMaterial(material, idx);
							}
							else
							{
								logError($"No skin part controller for material '{material.name}'");
							}
						}
					}
					else
					{
						logInfo("Character uses only one simple material. Transparency not possible");
					}

					shaderChanged = true;
					ForceUpdate();
				}
					
				foreach (var control in m_controls)
				{
					shaderChanged |= control.Value.ApplyTransparency();
				}
				if (shaderChanged)
				{
					// Hack to force a refresh of the shaders cache
					m_activeSkin?.BroadcastMessage("OnApplicationFocus", true);
				}
			}
		}

		protected void ForceUpdate()
		{
			foreach (var control in m_controls)
			{
				control.Value.ForceUpdate();
			}
		}

		protected void RestoreShaders()
		{
			foreach (var control in m_controls)
			{
				control.Value.RestoreShaders();
			}
			// Hack to force a refresh of the shaders cache
			m_activeSkin?.BroadcastMessage("OnApplicationFocus", true);
		}

		protected void logDebug(String _msg)
		{
			if (m_globalDebugOutput.val == true)
			{
				SuperController.LogMessage($"{nameof(GoingGoingGone)} ({m_person.name}): {_msg}");
			}
		}

		protected void logInfo(String _msg)
		{
			SuperController.LogMessage($"{nameof(GoingGoingGone)} ({m_person.name}): {_msg}");
		}

		protected void logError(String _msg)
		{
			SuperController.LogError($"{nameof(GoingGoingGone)} ({m_person.name}): {_msg}");
		}
	}
}